from .agent_conversation_history_settings import AgentConversationHistorySettings
from .agent_gatekeeper_settings import AgentGatekeeperSettings
from .agent_vectorization_settings import AgentVectorizationSettings
from .agent_base import AgentBase
from .knowledge_management_agent import KnowledgeManagementAgent
from .knowledge_management_completion_request import KnowledgeManagementCompletionRequest



from foundationallm.config import Configuration, Context
from foundationallm.models.orchestration import CompletionRequestBase
from foundationallm.langchain.agents import (
    LangChainAgentBase,
    LangChainKnowledgeManagementAgent
)

class AgentFactory:
    """
    Factory to determine which agent to use.
    """

    def __init__(
            self,
            completion_request: CompletionRequestBase,
            config: Configuration,
            context: Context
        ):
        """
        Initializes an Orchestration for selecting which agent to use for completion.

        Parameters
        ----------
        completion_request : CompletionRequest
            The completion request object containing the user prompt to execute, message history,
            and agent and data source metadata.
        config : Configuration
            Application configuration class for retrieving configuration settings.
        context : Context
            The user context under which to execution completion requests.    
        """
        self.completion_request = completion_request
        self.config = config
        self.context = context

    def get_agent(self) -> LangChainAgentBase:
        """
        Retrieves an agent of the the requested type.
        
        Returns
        -------
        AgentBase
            Returns an agent of the requested type.
        """
        agent = self.completion_request.agent
        if agent is None:
            raise ValueError("Agent not constructed. Cannot access an object of 'NoneType'.")
        
        match agent.type:
            case 'knowledge-management' | 'internal-context':
                return LangChainKnowledgeManagementAgent(config=self.config)
            case _:
                raise ValueError(f'The specified agent type {agent.type} is not supported.')



from io import BytesIO
import fnmatch
from azure.storage.blob import BlobServiceClient
from foundationallm.storage import StorageManagerBase
from azure.identity import DefaultAzureCredential

class BlobStorageManager(StorageManagerBase):
    """
    The BlobStorageManager class is responsible for managing files in Azure Blob Storage.
    """

    def __init__(self, blob_connection_string=None, container_name=None, account_name=None, authentication_type='ConnectionString'):
        """
        Initialize a blob storage manager.

        Parameters
        ----------
        blob_connection_string : str
            The connection string for the target blob storage account.
        container_name : str
            The name of the container is blob storage from which blobs should be retrieved.
        """
        if authentication_type == 'AzureIdentity':
            if account_name is None or account_name == '':
                raise ValueError('The account_name parameter must be set to a valid account name.')
            credential = DefaultAzureCredential(exclude_environment_credential=True)
            blob_service_client = BlobServiceClient(account_url=f"https://{account_name}.blob.core.windows.net", credential=credential)
        else:
            if blob_connection_string is None or blob_connection_string == '':
                raise ValueError('The blob_connection_string parameter must be set to a valid connection string.')
            blob_service_client = BlobServiceClient.from_connection_string(blob_connection_string)
            
        if container_name is None or container_name == '':
            raise ValueError('The container_name parameter must be set to a valid container.')
 
        self.blob_container_client = blob_service_client.get_container_client(container_name)

    def __get_full_path(self, path) -> str:
        """
        Retrieves the full path to the target blob.

        Parameters
        ----------
        path : str
            The path to the blob.

        Returns
        -------
        str
            The full path to the specified blob.
        """
        return '/'.join([i for i in path.split('/') if i != ''])

    def list_blobs(self, path, file_name_pattern=None):
        """
        Retrieves a list of blobs in the specified path. The path should be to a folder, not a file.

        Parameters
        ----------
        file_name_pattern : str
            Filter used to determine the results using a wildcard pattern for the file name.
            Example:
                file_name_pattern='*1980*.csv' returns all blobs that contain '1980' in the file name.

        Returns
        -------
        List[BlobProperties]
            A list of blobs in the specified path.
        """
        full_path = self.__get_full_path(path)
        blobs = self.blob_container_client.list_blobs(name_starts_with=full_path)

        if file_name_pattern is None:
            return blobs

        # Filter the blobs by the file name pattern and exclude any folders
        filtered_blobs = [blob for blob in blobs if fnmatch.fnmatch(blob.name, file_name_pattern)
                          and blob.content_settings.content_type is not None]

        return filtered_blobs

    def file_exists(self, path) -> bool:
        """
        Checks whether a specified path exists in blob storage.

        Parameters
        ----------
        path : str
            The path the check for existence.

        Returns
        -------
        bool
            Returns true of the specified path exists. Otherwise, false.
        """
        full_path = self.__get_full_path(path)
        blob = self.blob_container_client.get_blob_client(full_path)
        return blob.exists()

    def read_file_content(self, path, read_into_stream=True) -> bytes:
        """
        Retrieves the contents of a specified file in bytes.

        Parameters
        ----------
        path : str
            The path to the blob being retrieved.
        read_into_stream : boolean
            Flag indicating whether to read the file content into a byte stream. Default is True.

        Returns
        -------
        bytes
            Returns the bytes, or a stream of bytes, representing the content of the specified file
            or None if the file does not exist.
        """
        if self.file_exists(path):
            full_path = self.__get_full_path(path)
            if read_into_stream:
                blob = self.blob_container_client.get_blob_client(full_path)
                stream = BytesIO()
                blob.download_blob().readinto(stream)
                return stream.getvalue()
            else:
                blob = self.blob_container_client.download_blob(full_path)
                return blob.content_as_bytes()
        else:
            return None

    def write_file_content(self, path, content, overwrite=True, lease=None):
        """
        Writes data to a specified file.

        Parameters
        ----------
        path : str
            The path to the blob to which data is being written.
        content
            The data being written into the blob.
        overwrite : boolean
            Indicates whether the content of the blob should be overwritten by the incoming content.
            Default is True.
        lease
            The lease to use for accessing the blob. Default is None.
        """
        full_path = self.__get_full_path(path)
        blob = self.blob_container_client.get_blob_client(full_path)
        blob.upload_blob(content, overwrite=overwrite, lease=lease)

    def delete_file(self, path):
        """
        Deletes the specified blob.

        Parameters
        ----------
        path : str
            The path to the blob to be deleted.
        """
        full_path = self.__get_full_path(path)
        self.blob_container_client.delete_blob(full_path, delete_snapshots='include')



from typing import List, Optional, Literal
from foundationallm.langchain.data_sources import DataSourceConfiguration

class BlobStorageConfiguration(DataSourceConfiguration):
    """
    Connection information indicating the connection string to connect,
    the container to find the files, and the list of files to process.
    """
    configuration_type: Literal['blob_storage']
    connection_string_secret: str
    container: str
    files: Optional[List[str]] = None



from typing import Optional
from pydantic import BaseModel
from foundationallm.models.language_models import LanguageModelType, LanguageModelProvider

class EmbeddingModel(BaseModel):
    """Embedding model metadata model."""
    type: str = LanguageModelType.OPENAI
    provider: Optional[str] = LanguageModelProvider.MICROSOFT
    deployment: Optional[str] = "FoundationaLLM:AzureOpenAI:API:Embeddings:DeploymentName"    
    chunk_size: Optional[int] = 1000
    api_endpoint: Optional[str] = "FoundationaLLM:AzureOpenAI:API:Endpoint"
    api_key: Optional[str] = "FoundationaLLM:AzureOpenAI:API:Key"
    api_version: Optional[str] = "FoundationaLLM:AzureOpenAI:API:Version"



from abc import ABC, abstractmethod

from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseLanguageModel

from foundationallm.config import Configuration
from foundationallm.models.language_models import EmbeddingModel

class LanguageModelBase(ABC):
    """Abstract base class for language models."""

    def __init__(self, config: Configuration):
        """
        Initializer
        
        Parameters
        ----------
        app_config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.config = config

    @abstractmethod
    def get_embedding_model(self, embedding_model: EmbeddingModel) -> Embeddings:
        """
        Retrieve the embeddings model.
        
        Returns
        -------
        Embeddings
            The embeddings large language model to use.
        """



"""
Class: EmbeddingProfileBase
Description: Base class for vectorization embedding profiles.
"""
from foundationallm.models.resource_providers.vectorization import ProfileBase

class EmbeddingProfileBase(ProfileBase):
    """
    Base class to hold indexing profile information.
    """
    text_embedding : str



"""SQL database data source module."""
from .sql_database_configuration import SQLDatabaseConfiguration
from .sql_database_datasource import SQLDatabaseDataSource
from .mariadb import MariaDB
from .mssql import MicrosoftSQLServer
from .mysql import MySQL
from .postgresql import PostgreSQL
from .sql_database_factory import SQLDatabaseFactory



"""
Class: SearchServiceRetriever
Description: LangChain retriever for Azure AI Search.
"""
from typing import List, Optional

from langchain_openai import OpenAIEmbeddings , AzureOpenAIEmbeddings
from langchain_core.callbacks import (
    AsyncCallbackManagerForRetrieverRun,
    CallbackManagerForRetrieverRun,
)
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever

from azure.search.documents import SearchClient
from azure.search.documents.models import VectorizedQuery
from azure.core.credentials import AzureKeyCredential

class SearchServiceFilterRetriever(BaseRetriever):
    """
    LangChain retriever for Azure AI Search.
    Properties:
        endpoint: str -> Azure AI Search endpoint
        index_name: str -> Azure AI Search index name
        top_n : int -> number of results to return from vector search
        embedding_field_name: str -> name of the field containing the embedding vector
        text_field_name: str -> name of the field containing the raw text
        credential: AzureKeyCredential -> Azure AI Search credential
        embedding_model: OpenAIEmbeddings -> OpenAIEmbeddings model

    Searches embedding and text fields in the index for the top_n most relevant documents.

    Default FFLM document structure (overridable by setting the embedding and text field names):
        {
            "Id": "<GUID>",
            "Embedding": [0.1, 0.2, 0.3, ...], # embedding vector of the Text
            "Text": "text of the chunk",
            "Description": "General description about the source of the text",
            "AdditionalMetadata": "JSON string of metadata"
            "ExternalSourceName": "name and location the text came from, url, blob storage url"
            "IsReference": "true/false if the document is a reference document"
        }
    """
    endpoint: str
    index_name: str
    filters : List[str]
    top_n : int
    embedding_field_name: Optional[str] = "Embedding"
    text_field_name: Optional[str] = "Text"
    credential: AzureKeyCredential
    embedding_model: OpenAIEmbeddings

    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True

    def __get_embeddings(self, text: str) -> List[float]:
        """
        Returns embeddings vector for a given text.
        """
        embedding = self.embedding_model.embed_query(text)
        return embedding

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs a synchronous hybrid search on Azure AI Search index
        """

        results_list = []

        search_client = SearchClient(self.endpoint, self.index_name, self.credential)

        for filter in self.filters:

            try:
                vector_query = VectorizedQuery(vector=self.__get_embeddings(query),
                                                k_nearest_neighbors=3,
                                                fields=self.embedding_field_name)

                if (filter == "search.ismatch('*', 'metadata', 'simple', 'all')"):
                        results = search_client.search(
                        search_text=query,
                        vector_queries=[vector_query],
                        top=self.top_n,
                        select=[self.text_field_name]
                    )
                else:
                    results = search_client.search(
                        search_text=query,
                        filter=filter,
                        vector_queries=[vector_query],
                        top=self.top_n,
                        select=[self.text_field_name]
                    )

                for result in results:
                    try:
                        results_list.append(Document(
                            page_content=result[self.text_field_name]
                        ))
                    except Exception as e:
                        print(e)

            except Exception as e:
                print(e)

            if ( filter == "search.ismatch('*', 'metadata', 'simple', 'all')"):
                break

        return results_list

    async def _aget_relevant_documents(
        self, query: str, *, run_manager: AsyncCallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs an asynchronous hybrid search on Azure AI Search index
        NOTE: This functionality is not currently supported in the underlying Azure SDK.
        """
        raise Exception(f"Asynchronous search not supported.")



"""
Base Agent fulfilling the orchestration request.
"""
from pydantic import BaseModel
from typing import Optional
from foundationallm.models.orchestration import OrchestrationSettings
from foundationallm.models.agents import AgentConversationHistorySettings, AgentGatekeeperSettings

class AgentBase(BaseModel):
    """ Agent Base model."""
    name: str
    description: Optional[str]
    type: str
    object_id: Optional[str] = None
    prompt_object_id: Optional[str] = None
    sessions_enabled: Optional[bool] = False
    conversation_history: Optional[AgentConversationHistorySettings] = AgentConversationHistorySettings()
    gatekeeper: Optional[AgentGatekeeperSettings] = AgentGatekeeperSettings()
    orchestration_settings: Optional[OrchestrationSettings] = OrchestrationSettings()



from typing import List, Optional, Union
from pydantic import BaseModel
from .citation import Citation

class CompletionResponse(BaseModel):
    """
    Response from a language model.
    """
    user_prompt: str
    full_prompt: Optional[str] = None
    completion: Union[str, set, List[str]]
    citations: Optional[List[Citation]] = []
    user_prompt_embedding: Optional[List[float]] = []
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0
    total_cost: float = 0.0



from langchain_community.utilities.sql_database import SQLDatabase

from foundationallm.config import Configuration
from foundationallm.langchain.data_sources.sql import SQLDatabaseConfiguration
from foundationallm.langchain.data_sources.sql import MariaDB, MicrosoftSQLServer, MySQL, PostgreSQL
from .sql_database_dialects import SQLDatabaseDialect

class SQLDatabaseFactory():
    """
    Factory class to determine which SQL database to use.
    """
    def __init__(self, sql_db_config: SQLDatabaseConfiguration, config: Configuration):
        """
        Initializes an sql database factory.

        Parameters
        ----------
        sql_db_config : SQLDatabaseConfiguration
            SQL Database configuration class for providing settings to the database.
        config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.sql_db_config: SQLDatabaseConfiguration = sql_db_config
        self.config: Configuration = config
        self.dialect = self.sql_db_config.dialect

    def get_sql_database(self) -> SQLDatabase:
        match self.dialect:
            case SQLDatabaseDialect.MARIADB:
                return MariaDB(sql_db_config = self.sql_db_config,
                               config = self.config).get_database()
            case SQLDatabaseDialect.MSSQL:
                return MicrosoftSQLServer(sql_db_config = self.sql_db_config,
                                          config = self.config).get_database()
            case SQLDatabaseDialect.MYSQL:
                return MySQL(sql_db_config = self.sql_db_config,
                             config = self.config).get_database()
            case SQLDatabaseDialect.POSTGRESQL:
                return PostgreSQL(sql_db_config = self.sql_db_config,
                                  config = self.config).get_database()



from typing import List, Optional, Literal
from foundationallm.langchain.data_sources import DataSourceConfiguration

class SQLDatabaseConfiguration(DataSourceConfiguration):
    """Configuration settings for a connection to a SQL database."""
    configuration_type: Literal['sql_database']
    dialect: str
    host: str
    port: int = None
    database_name: str
    username: str = None
    password_secret_setting_key_name: str = None
    include_tables: Optional[List[str]] = None
    exclude_tables: Optional[List[str]] = None
    few_shot_example_count: int = 0
    row_level_security_enabled: bool = False



from urllib import parse

from foundationallm.langchain.data_sources.sql import SQLDatabaseDataSource

class MySQL(SQLDatabaseDataSource):
    """
    Generates a MySQL database connection.
    """
    def get_connection_string(self) -> str:
        """
        Gets a formatted connection string for connecting to the database.
        
        Returns
        -------
        int
            Returns a formatted connection string for connecting to the database.
        """
        return f'{self.dialect}+{self.driver}://{self.username}:{parse.quote_plus(self.password)}@{self.host}/{self.database_name}?charset=utf8mb4'

    def get_driver(self) -> str:
        """
        Gets the driver for connecting to the database.
        
        Returns
        -------
        str
            Returns the driver for connecting to the database.
        """
        return 'pymysql'

    def get_default_port(self) -> int:
        """
        Gets the default port of the database.
        
        Returns
        -------
        int
            Returns the default port for the database.
        """
        return 3306



"""
Class: SettingsBase
Description: Base settings for a vectorization profile
"""
from pydantic import BaseModel

class SettingsBase(BaseModel):
    """
    Base settings for a vectorization profile.
    """
    pass



"""
Class: AzureAISearchServiceRetriever
Description: LangChain retriever for Azure AI Search.
"""
import json
from typing import List, Optional, Union, Tuple
from langchain_openai import OpenAIEmbeddings
from langchain_core.callbacks import (
    AsyncCallbackManagerForRetrieverRun,
    CallbackManagerForRetrieverRun,
)
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from azure.search.documents import SearchClient
from azure.search.documents.models import VectorizedQuery
from azure.core.credentials import AzureKeyCredential
from azure.identity import DefaultAzureCredential
from foundationallm.models.orchestration import Citation
from .citation_retrieval_base import CitationRetrievalBase

class AzureAISearchServiceRetriever(BaseRetriever, CitationRetrievalBase):
    """
    LangChain retriever for Azure AI Search.
    Properties:
        endpoint: str -> Azure AI Search endpoint
        index_name: str -> Azure AI Search index name
        top_n : int -> number of results to return from vector search
        embedding_field_name: str -> name of the field containing the embedding vector
        text_field_name: str -> name of the field containing the raw text
        id_field_name: str -> name of the field containing the document unique identifier in the index
        metadata_field_name -> str -> name of the field containing the JSON metadata as a string
        filters: str -> Azure AI Search filter expression
        credential: AzureKeyCredential -> Azure AI Search credential
        embedding_model: OpenAIEmbeddings -> OpenAIEmbeddings model

    Searches embedding and text fields in the index for the top_n most relevant documents.

    Default FFLM document structure (overridable by setting the embedding and text field names):
        {
            "Id": "<GUID>",
            "Embedding": [0.1, 0.2, 0.3, ...], # embedding vector of the Text
            "Text": "text of the chunk",
            "Description": "General description about the source of the text",            
            "AdditionalMetadata": "JSON string of metadata"
            "ExternalSourceName": "name and location the text came from, url, blob storage url"
            "IsReference": "true/false if the document is a reference document"
        }
    """
    endpoint: str
    index_name: str
    top_n : int
    embedding_field_name: Optional[str] = "Embedding"
    text_field_name: Optional[str] = "Text"
    id_field_name: Optional[str] = "Id"
    metadata_field_name: Optional[str] = "AdditionalMetadata"
    filters: Optional[str] = None
    credential: Union[AzureKeyCredential, DefaultAzureCredential] = None
    embedding_model: OpenAIEmbeddings
    search_results: Optional[Tuple[str, Document]] = [] # Tuple of document id and document

    def __get_embeddings(self, text: str) -> List[float]:
        """
        Returns embeddings vector for a given text.
        """
        embedding = self.embedding_model.embed_query(text)        
        return embedding

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs a synchronous hybrid search on Azure AI Search index        
        """        
        search_client = SearchClient(self.endpoint, self.index_name, self.credential)        
        vector_query = VectorizedQuery(vector=self.__get_embeddings(query),
                                        k_nearest_neighbors=3,
                                        fields=self.embedding_field_name)

        results = search_client.search(
            search_text=query,
            filter=self.filters,
            vector_queries=[vector_query],
            top=self.top_n,
            select=[self.id_field_name, self.text_field_name, self.metadata_field_name]
        )
        self.search_results.clear()
        for result in results:
            metadata = json.loads(result[self.metadata_field_name]) if self.metadata_field_name in result else {}
            document = Document(
                    page_content=result[self.text_field_name],
                    metadata=metadata                                        
            )
            self.search_results.append((result[self.id_field_name], document))
        return [doc for _, doc in self.search_results]

    async def _aget_relevant_documents(
        self, query: str, *, run_manager: AsyncCallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Performs an asynchronous hybrid search on Azure AI Search index
        NOTE: This functionality is not currently supported in the underlying Azure SDK.
        """
        raise Exception(f"Asynchronous search not supported.")

    def get_document_citations(self) -> List[Citation]:
        """
        Gets sources from the documents retrieved from the retriever.
  
        Returns:
            List of citations from the retrieved documents.
        """
        citations = []
        added_ids = set()  # Avoid duplicates
        for result_id, result in self.search_results:  # Unpack the tuple
            metadata = result.metadata
            if metadata is not None and 'multipart_id' in metadata and metadata['multipart_id']:
                if result_id not in added_ids:          
                    title = (metadata['multipart_id'][-1]).split('/')[-1]
                    filepath = '/'.join(metadata['multipart_id'])
                    citations.append(Citation(id=result_id, title=title, filepath=filepath))
                    added_ids.add(result_id)
        return citations

    def format_docs(self, docs:List[Document]) -> str:
        """
        Generates a formatted string from a list of documents for use
        as the context for the completion request.
        """
        return "\n\n".join(doc.page_content for doc in docs)



from enum import Enum

class SQLDatabaseDialect(str, Enum):
    """Enumerator of SQL database dialects"""
    MARIADB = "mariadb"
    MSSQL = "mssql"
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"



import logging
from azure.monitor.opentelemetry import configure_azure_monitor
from opentelemetry import trace
from opentelemetry.trace import Span, Status, StatusCode, Tracer
from foundationallm.config import Configuration

class Telemetry:
    """
    Manages logging and the recording of application telemetry.
    """

    @staticmethod
    def configure_monitoring(config: Configuration, telemetry_connection_string: str):
        """
        Configures monitoring and sends logs, metrics, and events to Azure Monitor.

        Parameters
        ----------
        config : Configuration
            Configuration class used for retrieving application settings from
            Azure App Configuration.
        telemetry_connection_string : str
            The connection string used to connect to Azure Application Insights.
        """
        configure_azure_monitor(
            connection_string=config.get_value(telemetry_connection_string),
            disable_offline_storage=True
        )

    @staticmethod
    def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:
        """
        Creates a logger by the specified name and logging level.

        Parameters
        ----------
        name : str
            The name to assign to the logger instance.
        level : int
            The logging level to assign.

        Returns
        -------
        Logger
            Returns a logger object with the specified name and logging level.
        """
        logger = logging.getLogger(name)
        logger.setLevel(level)
        return logger

    @staticmethod
    def get_tracer(name: str) -> Tracer:
        """
        Creates an OpenTelemetry tracer with the specified name.

        Parameters
        ----------
        name : str
            The name to assign to the tracer.

        Returns
        -------
        Tracer
            Returns an OpenTelemetry tracer for span creation and in-process context propagation.
        """
        return trace.get_tracer(name)

    @staticmethod
    def record_exception(span: Span, ex: Exception):
        """
        Associates an exception with an OpenTelemetry Span and logs it.

        Parameters
        ----------
        span : Span
            The OpenTelemetry span to which the execption should be associated.
        ex : Exception
            The exception that occurred.
        """
        span.set_status(Status(StatusCode.ERROR))
        span.record_exception(ex)



from typing import Optional
from pydantic import BaseModel

class OrchestrationSettings(BaseModel):
    """
        Represents settings for the orchestration.
        
        Parameters
        ----------
        agent_parameters : dict
            Parameters to set or override the behavior of the agent.
        endpoint_configuration : dict
            Parameters to set or override the endpoint configuration used by the orchestrator
            to access a language model.
        model_parameters : dict
            Parameters to set or override the behavior of the language model as defined on the agent.
        orchestrator : str
            The orchestrator to use for the orchestration. Defaults to 'LangChain'.
    """
    agent_parameters: Optional[dict] = None
    endpoint_configuration: Optional[dict] = None
    model_parameters: Optional[dict] = None
    orchestrator: Optional[str] = 'LangChain'



import re
from abc import abstractmethod
from typing import List, Optional
from azure.identity import DefaultAzureCredential, get_bearer_token_provider

from langchain_core.language_models import BaseLanguageModel
from langchain_openai import AzureChatOpenAI, AzureOpenAI, ChatOpenAI, OpenAI
from foundationallm.config.configuration import Configuration
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.models.orchestration import OperationTypes
from foundationallm.models.agents import AgentBase
from foundationallm.models.authentication import AuthenticationTypes
from foundationallm.models.language_models import LanguageModelProvider
from foundationallm.models.orchestration import (
    CompletionRequestBase,
    CompletionResponse,
    EndpointSettings,
    MessageHistoryItem,
    OrchestrationSettings
)
from foundationallm.models.resource_providers.prompts.multipart_prompt import MultipartPrompt
from foundationallm.models.resource_providers.vectorization import (
    AzureAISearchIndexingProfile,
    AzureOpenAIEmbeddingProfile
)

class LangChainAgentBase():
    """
    Implements the base functionality for a LangChain agent.
    """
    def __init__(self, config: Configuration):
        """
        Initializes a knowledge management agent.

        Parameters
        ----------
        config : Configuration
            Application configuration class for retrieving configuration settings.
        """
        self.config = config
        self.full_prompt = ""

    @abstractmethod
    def invoke(self, request: CompletionRequestBase) -> CompletionResponse:
        """
        Gets the completion for the request.
        
        Parameters
        ----------
        request : CompletionRequestBase
            The completion request to execute.

        Returns
        -------
        CompletionResponse
            Returns a completion response.
        """
        raise NotImplementedError()

    def _get_prompt_from_object_id(self, prompt_object_id: str, agent_parameters: dict) -> MultipartPrompt:
        """
        Get the prompt from the object id.
        """
        prompt: MultipartPrompt = None

        if prompt_object_id is None or prompt_object_id == '':
            raise LangChainException("Invalid prompt object id.", 400)
        
        try:
            prompt = MultipartPrompt(**agent_parameters.get(prompt_object_id))
        except Exception as e:
            raise LangChainException(f"The prompt object provided in the agent parameters is invalid. {str(e)}", 400)
        
        if prompt is None:
            raise LangChainException("The prompt object is missing in the agent parameters.", 400)

        return prompt

    def _validate_request(self, request: CompletionRequestBase):
        """
        Validates that the completion request's agent contains all require properties.

        Parameters
        ----------
        request : CompletionRequestBase
            The completion request to validate.
        """
        if request.agent is None:
            raise LangChainException("The Agent property of the completion request cannot be null.", 400)
        
        if request.agent.orchestration_settings is None:
            raise LangChainException("The OrchestrationSettings property of the agent cannot be null.", 400)
        
        if request.agent.orchestration_settings.endpoint_configuration is None:
            raise LangChainException("The EndpointConfiguration property of the OrchestrationSettings cannot be null.", 400)
        
        if request.agent.orchestration_settings.endpoint_configuration.get('endpoint') is None:
            raise LangChainException("The Endpoint property of the agent's OrchestrationSettings.EndpointConfiguration property cannot be null.", 400)

        autentication_type = request.agent.orchestration_settings.endpoint_configuration.get('auth_type')
        if autentication_type is None:
            raise LangChainException("The AuthType property of the agent's OrchestrationSettings.EndpointConfiguration property cannot be null.", 400)

        try:
            AuthenticationTypes(autentication_type)
        except ValueError:
            raise LangChainException(f"The authentication type {autentication_type} is not supported.", 400)
        
        provider = request.agent.orchestration_settings.endpoint_configuration.get('provider')
        if provider is None:
            raise LangChainException("The Provider property of the agent's OrchestrationSettings.EndpointConfiguration property cannot be null.", 400)

        try:
            LanguageModelProvider(provider)
        except ValueError:
            raise LangChainException(f"The LLM provider {provider} is not supported.", 400)

        if provider == LanguageModelProvider.MICROSOFT:
            # Verify the endpoint_configuration inludes the api_version property for Azure OpenAI models.
            if request.agent.orchestration_settings.endpoint_configuration.get('api_version') is None:
                raise LangChainException("The ApiVersion property of the agent's OrchestrationSettings.EndpointConfiguration property cannot be null.", 400)
            
            # model_parameters is required to provide the deployment_name for Azure OpenAI models.
            if request.agent.orchestration_settings.model_parameters is None:
                raise LangChainException("The ModelParameters property of the OrchestrationSettings cannot be null.", 400)

            # Verify that the deployment_name is provided for Azure OpenAI models.
            if request.agent.orchestration_settings.model_parameters.get('deployment_name') is None:
                raise LangChainException("The DeploymentName property of the agent's OrchestrationSettings.ModelParameters property cannot be null.", 400)

    def _build_conversation_history(self, messages:List[MessageHistoryItem]=None, message_count:int=None) -> str:
        """
        Builds a chat history string from a list of MessageHistoryItem objects to
        be added to the prompt for the completion request.

        Parameters
        ----------
        messages : List[MessageHistoryItem]
            The list of messages from which to build the chat history.
        message_count : int
            The number of messages to include in the chat history.
        """
        if messages is None or len(messages)==0:
            return ""
        if message_count is not None:
            messages = messages[-message_count:]
        chat_history = "Chat History:\n"
        for msg in messages:
            chat_history += msg.sender + ": " + msg.text + "\n"
        chat_history += "\n\n"
        return chat_history

    def _record_full_prompt(self, prompt: str) -> str:
        """
        Records the full prompt for the completion request.

        Parameters
        ----------
        prompt : str
            The prompt that is populated with context.
        
        Returns
        -------
        str
            Returns the full prompt.
        """
        self.full_prompt = prompt
        return prompt

    def __extract_endpoint_configuration(
            self,
            endpoint_configuration: dict) -> EndpointSettings:
        """
        Extracts the endpoint configuration settings from the agent's orchestration settings.

        Parameters
        ----------
        config : Configuration
            Application configuration class for retrieving configuration settings.
        endpoint_configuration : dict
            The endpoint configuration settings to extract.

        Returns
        -------
        EndpointSettings
            Returns the endpoint settings for the completion request.
        """
        # TODO: Also allow for the override of the endpoint configuration?
        endpoint_settings = EndpointSettings(
            endpoint=endpoint_configuration.get('endpoint'),
            authentication_type=endpoint_configuration.get('auth_type'),
            provider=endpoint_configuration.get('provider'),
            api_version=endpoint_configuration.get('api_version'),
            operation_type=endpoint_configuration.get('operation_type') or 'chat'
        )

        endpoint_settings.api_type = 'azure_ad' if endpoint_settings.authentication_type == 'token' else 'azure'

        if endpoint_settings.authentication_type == AuthenticationTypes.KEY:
            endpoint_settings.api_key = self.config.get_value(endpoint_configuration.get('api_key'))
            if endpoint_settings.api_key is None:
                raise ValueError(f"API Key is required for completion requests using {endpoint_settings.authentication_type}-based authentication.")

        return endpoint_settings
        

    def _get_language_model(
            self,
            agent_orchestration_settings: OrchestrationSettings,
            model_override_settings: Optional[OrchestrationSettings] = None) -> BaseLanguageModel:
        """
        Create a language model using the specified endpoint settings.

        Parameters
        ----------
        config : Configuration
            Application configuration class for retrieving configuration settings.
        agent_orchestration_settings : OrchestrationSettings
            The settings for the completion request configured on the agent.
        model_override_settings : Optional[OrchestrationSettings]
            Any settings specified on the completion request for overriding the model's settings.

        Returns
        -------
        BaseLanguageModel
            Returns an API connector for a chat completion model.
        """
        endpoint_settings = self.__extract_endpoint_configuration(agent_orchestration_settings.endpoint_configuration)
                
        langauge_model:BaseLanguageModel = None

        if endpoint_settings.provider == LanguageModelProvider.MICROSOFT:
            # Get Azure OpenAI Chat model settings
            deployment_name = (model_override_settings.model_parameters.get('deployment_name')
                                if model_override_settings is not None
                                    and model_override_settings.model_parameters is not None
                                    and model_override_settings.model_parameters.get('deployment_name') is not None
                                else agent_orchestration_settings.model_parameters.get('deployment_name'))
            if deployment_name is None:
                raise ValueError("Deployment name is required for Azure OpenAI completion requests.")

            if endpoint_settings.authentication_type == AuthenticationTypes.TOKEN:
                try:
                    # Set up a Azure AD token provider.
                    # TODO: Determine if there is a more efficient way to get the token provider than making the request for every call.
                    token_provider = get_bearer_token_provider(
                        DefaultAzureCredential(exclude_environment_credential=True),
                        'https://cognitiveservices.azure.com/.default'
                    )
                
                    langauge_model = (
                        AzureChatOpenAI(
                            azure_endpoint=endpoint_settings.endpoint,
                            api_version=endpoint_settings.api_version,
                            openai_api_type=endpoint_settings.api_type,
                            azure_ad_token_provider=token_provider,
                            azure_deployment=deployment_name
                        ) if endpoint_settings.operation_type == OperationTypes.CHAT
                        else AzureOpenAI(
                            azure_endpoint=endpoint_settings.endpoint,
                            api_version=endpoint_settings.api_version,
                            openai_api_type=endpoint_settings.api_type,
                            azure_ad_token_provider=token_provider,
                            azure_deployment=deployment_name
                        )
                    )
                except Exception as e:
                    raise LangChainException(f"Failed to create Azure OpenAI API connector: {str(e)}", 500)
            else: # Key-based authentication
                langauge_model = (
                    AzureChatOpenAI(
                        azure_endpoint=endpoint_settings.endpoint,
                        api_key=endpoint_settings.api_key,
                        api_version=endpoint_settings.api_version,
                        azure_deployment=deployment_name
                    ) if endpoint_settings.operation_type == OperationTypes.CHAT
                    else AzureOpenAI(
                        azure_endpoint=endpoint_settings.endpoint,
                        api_key=endpoint_settings.api_key,
                        api_version=endpoint_settings.api_version,
                        azure_deployment=deployment_name
                    )
                )
        else:
            langauge_model = (
                ChatOpenAI(base_url=endpoint_settings.endpoint, api_key=endpoint_settings.api_key)
                if endpoint_settings.operation_type == OperationTypes.CHAT
                else OpenAI(base_url=endpoint_settings.endpoint, api_key=endpoint_settings.api_key)
            )

        # Set model parameters from agent orchestration settings.
        for key, value in agent_orchestration_settings.model_parameters.items():
            if hasattr(langauge_model, key):
                setattr(langauge_model, key, value)

        # Override model parameters from completion request settings, if any exist.
        if model_override_settings is not None and model_override_settings.model_parameters is not None:            
            for key, value in model_override_settings.model_parameters.items():
                if hasattr(langauge_model, key):
                    setattr(langauge_model, key, value)

        return langauge_model



from pydantic import BaseModel

class MessageHistoryItem(BaseModel):
    """
        Represents an historic message sender and text item.
        
        Parameters
        ----------
        sender : str
            The sender of the message (e.g., "Agent", "User")
        text : str
            The message text.
    """
    sender: str
    text: str



from pydantic import BaseModel
from typing import Optional

from .operation_types import OperationTypes
from foundationallm.models.authentication import AuthenticationTypes

class EndpointSettings(BaseModel):
    endpoint: str
    api_key: Optional[str] = None
    api_type: Optional[str] = None
    api_version: Optional[str] = None
    authentication_type: str = AuthenticationTypes.TOKEN
    operation_type: str = OperationTypes.CHAT
    provider: str



from typing import Optional
from pydantic import BaseModel

class Citation(BaseModel):
    """
    Source reference information for a completion
    """
    id: str # id in the index
    title: Optional[str] = None # file name
    filepath: Optional[str] = None



"""
Classes:
    - SettingsBase: Base settings for a vectorization profile
    - ConfigurationReferencesBase: Base class to hold configuration references for a vectorization profile
    - ProfileBase: Base class to hold vectorization profile information
Description: Base class to hold vectorization profile information
"""
from typing import Optional
from pydantic import BaseModel
from .settings_base import SettingsBase
from .configuration_references_base import ConfigurationReferencesBase

class ProfileBase(BaseModel):
    """
    Base class to hold vectorization profile information.
    """
    name: str
    object_id: Optional[str] = None
    settings: SettingsBase
    configuration_references: ConfigurationReferencesBase
